import os.path as osp
import os
import sys
sys.path.append(os.getcwd())
import torch
from torch.nn import Linear
import torch.nn.functional as F
from torch_geometric.datasets import Planetoid
import torch_geometric.transforms as T
from torch_geometric.nn.conv.gcn_conv import gcn_norm
from torch.nn import Sequential as Seq, Linear as Lin, ReLU, BatchNorm1d as BN, Conv1d, LeakyReLU, Conv2d, \
    BatchNorm2d as BN2D
from torch.nn import Sequential as Seq, Dropout, Linear as Lin
from torch_cluster import random_walk
from torch_scatter import scatter_add
num_workers = 6
gettrace = getattr(sys, 'gettrace', None)
if gettrace is not None:
    num_workers = 0


def MLP(channels, batch_norm=True):
    return Seq(*[
        Seq(Lin(channels[i - 1], channels[i]))
        for i in range(1, len(channels))
    ])


def conv2d(channels, kernelSize=3, depthWise=False):
    numgroups = 1 if not depthWise else channels[-1]
    return Seq(*[
        Seq(Conv2d(in_channels=channels[i - 1], out_channels=channels[i], kernel_size=(1, kernelSize),
                   padding=(0, kernelSize // 2), padding_mode='circular', groups=numgroups),
            )
        for i in range(1, len(channels))
    ])


class PathConv(torch.nn.Module):
    def __init__(self, num_channels, layerNumber,
                 num_paths=5, pathLength=5, depthwise=True,
                 learnSingleSpatConv=False, globalConv=None):
        super(PathConv, self).__init__()
        self.learnSingleSpatConv = learnSingleSpatConv
        self.globalConv = globalConv
        self.num_channels = num_channels
        self.num_paths = num_paths
        self.pathLength = pathLength
        self.depthwise = depthwise
        self.conv1 = conv2d([self.num_channels, self.num_channels], kernelSize=self.pathLength,
                            depthWise=self.depthwise)
        if self.learnSingleSpatConv:
            self.spatConv = torch.nn.Parameter(1e-1 * torch.randn((1, self.pathLength)).cuda())
        self.conv1_1 = MLP([self.num_channels, self.num_channels])

    def forward(self, x, x_0, data, node_weights=None, xprev=None):
        num_verts = x.shape[0]
        edge_index = data.edge_index
        nodesIdx = torch.arange(num_verts, device=edge_index.device).repeat_interleave(self.num_paths)
        w = random_walk(edge_index[0, :], edge_index[1, :], start=nodesIdx, walk_length=self.pathLength - 1,
                        p=1, q=1, num_nodes=x.shape[0])
        # Reshape to [N, pathLength, num_paths]
        w = w.view(num_verts, self.num_paths, self.pathLength).transpose(1, 2)
        if node_weights is not None:
            weights = node_weights[w]
            weights_cumprod = torch.cumprod(weights, dim=1)
            x = (weights_cumprod[:, :, :, None] * x[w, :]).transpose(1, -1)
        else:
            x = x[w, :].transpose(1, -1)
        if not self.depthwise:
            x = self.conv1(x)
            x = x[:, :, :, 0]
            x = F.relu(x)
        else:
            if self.globalConv is not None:
                K = torch.repeat_interleave(self.globalConv, self.num_channels, dim=0).unsqueeze(1)
                x = (K * x).sum(dim=-1)
            elif self.learnSingleSpatConv:
                K = torch.repeat_interleave(self.spatConv, self.num_channels, dim=0).unsqueeze(1)
                x = (K * x).sum(dim=-1)
            else:
                x = (self.conv1[0][0].weight.squeeze(1) * x).sum(dim=-1)
        x = F.avg_pool1d(x, kernel_size=self.num_paths).squeeze()
        x = self.conv1_1(x)
        return x


class PathGCN(torch.nn.Module):
    def __init__(self, num_features, num_classes, hidden_channels, num_layers,
                 dropout=0.0, num_paths=5, pathLength=5, depthwise=True, learnSingleSpatConv=False,
                 globalConv=False):
        super(PathGCN, self).__init__()
        self.globalConv = None
        if globalConv:
            self.globalConv = torch.nn.Parameter(1e-2 * torch.randn((1, pathLength)).cuda())

        self.learnSingleSpatConv = learnSingleSpatConv
        self.depthwise = depthwise
        self.num_paths = num_paths
        self.pathLength = pathLength
        self.num_in_features = num_features
        self.num_classes = num_classes
        self.lins = torch.nn.ModuleList()
        self.lins.append(Linear(num_features, hidden_channels))
        self.lins.append(Linear(hidden_channels, num_classes))
        self.convs = torch.nn.ModuleList()
        self.spatialConvs = torch.nn.ModuleList()
        self.convs1x1 = torch.nn.ModuleList()

        self.singleSpatConvs = torch.nn.ParameterList()  # []
        self.dropout = dropout
        for layer in range(num_layers):
            self.convs.append(
                PathConv(hidden_channels, layer + 1,
                         num_paths=self.num_paths,
                         pathLength=self.pathLength,
                         depthwise=self.depthwise,
                         learnSingleSpatConv=self.learnSingleSpatConv,
                         globalConv=self.globalConv))
            self.spatialConvs.append(self.convs[-1].conv1)
            self.convs1x1.append(self.convs[-1].conv1_1)
            if self.learnSingleSpatConv:
                self.singleSpatConvs.append(self.convs[-1].spatConv)

        self.dropout = dropout

    def forward(self, x, data):
        num_verts = x.shape[0]
        row, col = data.edge_index[0], data.edge_index[1]
        deg = scatter_add(data.edge_weights, col, dim=0, dim_size=num_verts)
        deg_inv_sqrt = deg.pow_(-0.5)
        deg_inv_sqrt.masked_fill_(deg_inv_sqrt == float('inf'), 0)
        x = F.dropout(x, self.dropout, training=self.training)
        x = x_0 = self.lins[0](x).relu()
        xprev = x.clone()
        for conv in self.convs:
            xn_tmp = x.clone()
            x = F.dropout(x, self.dropout, training=self.training)
            x = F.relu(conv(x, x_0, data, node_weights=deg_inv_sqrt, xprev=xprev))
            xprev = xn_tmp

        x = F.dropout(x, self.dropout, training=self.training)
        x = self.lins[1](x)
        return x.log_softmax(dim=-1)


###################################################################################
num_layers = 2
dataset = 'Cora'
if dataset == 'Cora':
    nNin = 1433
    n_channels = 64
    dropout = 0.6
    lamda = 0.5


elif dataset == 'CiteSeer':
    nNin = 3703
    n_channels = 256
    dropout = 0.7
    lamda = 0.6


elif dataset == 'PubMed':
    nNin = 500
    n_channels = 256
    dropout = 0.5
    lamda = 0.4

nopen = n_channels
nhid = n_channels
nNclose = n_channels
path = 'mypath/' + dataset
transform = T.Compose([T.NormalizeFeatures()])
dataset = Planetoid(path, dataset, transform=transform)
data = dataset[0]
[data.edge_index, data.edge_weights] = gcn_norm(data.edge_index)
num_nodes = data.x.shape[0]
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
data = data.to(device)
lr = 0.01
lrGCN = 0.001
wd = 1e-5
wdGCN = 1e-5
pathLength = 5
num_paths = 10
model = PathGCN(nNin, dataset.num_classes, hidden_channels=n_channels, num_layers=2, dropout=dropout,
                num_paths=num_paths,
                pathLength=pathLength,
                depthwise=True,
                learnSingleSpatConv=False, globalConv=False)
optimizer = torch.optim.Adam([
    dict(params=model.lins.parameters(), weight_decay=wd),
    dict(params=model.convs.parameters(), lr=lrGCN, weight_decay=wdGCN),
], lr=lr)
model.to(device)


def train():
    model.train()
    optimizer.zero_grad()
    out = model(data.x, data)
    loss = F.nll_loss(out[data.train_mask], data.y[data.train_mask])
    loss.backward()
    optimizer.step()
    return float(loss)


@torch.no_grad()
def eval_test():
    model.eval()
    out = model(data.x, data)
    pred, accs = out.argmax(dim=-1), []
    with torch.no_grad():
        test_preds = out[data['test_mask'], :]

    for _, mask in data('train_mask', 'val_mask', 'test_mask'):
        accs.append(int((pred[mask] == data.y[mask]).sum()) / int(mask.sum()))
    return accs, test_preds


best_val_acc = test_acc = 0
patience = 0
for epoch in range(1, 10001):
    debug = False
    loss = train()
    tmptmp_test_acc = 0
    all_test_accs = []
    best_val_tmp = 0
    for tidx in range(20):
        (train_acc, val_acc, tmp_test_acc), test_preds = eval_test()
        if val_acc > best_val_tmp:
            best_val_tmp = val_acc
            tmptmp_test_acc = tmp_test_acc
        all_test_accs.append(tmp_test_acc)
    all_test_accs = torch.tensor(all_test_accs)
    tmp_test_acc = tmptmp_test_acc

    if best_val_tmp > best_val_acc:
        best_val_acc = best_val_tmp
        test_acc = tmp_test_acc
        print(f'Epoch: {epoch:04d}, Loss: {loss:.4f} Train: {train_acc:.4f}, '
              f'Val: {val_acc:.4f}, Test: {tmp_test_acc:.4f}, '
              f'Final Test: {test_acc:.4f}', flush=True)
        print("MEAN / STD of tests:", torch.mean(100 * all_test_accs), ' / ', torch.std(100 * all_test_accs),
              flush=True)
        patience = 0
    else:
        patience = patience + 1
    if patience > 300:
        break
